% Main Program
% 
% Instructions:
%
% 1) Users have to choose a zero tolerance. Any values that are in magnitude
%    less than that value are numerically zero.
% 2) Users have to specify the objective ODE and the initial condition.
% 3) Users have to specify the stepsize. The program currently supports only
%    equidistant time-stepping.
% 4) Users have to specify a viewing window [0,T] in time and [ymin,ymax] for
%    the observed range of values of the solution.
% 5) Users have to assign whether to carry out integrity checks on the stepsize
%    for the method at run-time at each step, or determine a valid stepsize
%    prior to running the program. This is the purpose of boolean apriori:
%    apriori=true means no integrity checks at runtime.
%    apriori=false means the program will check the integrity of the stepsize
%    at runtime at each step
% 6) The program calls the function Arccot, provided separately. Arccot is the
%    real inverse cotangent function with range (0,pi).

%
% User specifications:
%

% Set zero tolerance
zero=1e-14;
% Use symbolic engine for y
syms y 
% Define objective ODE
dydt=exp(y);
% Set initial value
y0=2;
% Set stepsize
h=0.01;
% Set viewing window
ymin=0;
ymax=2.01;
T=2; %time interval [0,T]
% Assign true or false to apriori
apriori=false;


% Initialize the row matrix t. t contains the time steps.
t=0:h:T; 
% Initialize the row matrix yNumerical. yNumerical will later contain the
% numerical approximations for the solution on the time grid.
yNumerical=zeros(1,length(t));
% Store the initial value into the matrix yNumerical
yNumerical(1)=y0;  
for i=1:length(t)-1 
    a=double(subs(diff(dydt,2),y,yNumerical(i))/2);  
    b=double(subs(diff(dydt,1),y,yNumerical(i)));
    c=double(subs(dydt,y,yNumerical(i)));
    delta=b^2-4*a*c;
    stabilityCheck=2-h*b;
    if apriori==true
       if delta>=4*zero
          w=2*c*sinh(sqrt(delta)*h/2)/(sqrt(delta)*cosh(sqrt(delta)*h/2)-b*sinh(sqrt(delta)*h/2));
       elseif delta<=-4*zero
          w=2*c*sin(sqrt(-delta)*h/2)/(sqrt(-delta)*cos(sqrt(-delta)*h/2)-b*sin(sqrt(-delta)*h/2));
       elseif abs(delta)<4*zero
          w=2*c*h/(2-b*h)-h^3*c*delta/(3*(2-b*h)^2);
       end  
    elseif apriori==false
           if stabilityCheck<sqrt(zero)
              warning('Method requires a smaller stepsize in order to be stable.')
              t=t(1,1:i);
              yNumerical=yNumerical(1,1:i);
              break
           else
              if delta>=4*zero
                 if sqrt(delta)<b
                    hmax=(1/sqrt(delta))*log((b+sqrt(delta))/(b-sqrt(delta)));
                 else
                    hmax=inf;
                 end
                 if h<hmax
                    w=2*c*sinh(sqrt(delta)*h/2)/(sqrt(delta)*cosh(sqrt(delta)*h/2)-b*sinh(sqrt(delta)*h/2));
                 else
                    t=t(1,1:i);
                    yNumerical=yNumerical(1:1:i); 
                    Warn=['The algorithm terminates at step ',num2str(i-1),', approximating the solution on the interval [0,',num2str((i-1)*h),'], because the method becomes undefined in the next step due to the chosen stepsize. Suggestion: Rerun the program with smaller stepsize'];
                    warning(Warn)
                    break
                 end
              elseif delta<=-4*zero
                     hmax=(2/sqrt(-delta))*Arccot(b/sqrt(-delta));
                     if h<hmax
                        w=2*c*sin(sqrt(-delta)*h/2)/(sqrt(-delta)*cos(sqrt(-delta)*h/2)-b*sin(sqrt(-delta)*h/2));
                     else
                        Warn=['The algorithm terminates at step ',num2str(i-1),', approximating the solution on the interval [0,',num2str((i-1)*h),'], because the method becomes undefined in the next step due to the chosen stepsize. Suggestion: Rerun the program with smaller stepsize'];
                        warning(Warn)
                        t=t(1,1:i);
                        yNumerical=yNumerical(1,1:i); 
                        break
                     end
              elseif abs(delta)<4*zero
                     w=2*c*h/(2-b*h)-h^3*c*delta/(3*(2-b*h)^2);
              end  
           end
    end
    yassume=w+yNumerical(i);
    if yassume>ymax||yassume<ymin
       Warn=['The algorithm terminates at step ',num2str(i-1),', approximating the solution on the interval [0,',num2str((i-1)*h),'], because the approximate solution leaves the tracking window in the next step. Suggestion: Rerun the program with larger y-viewing window.'];
       warning(Warn)
       t=t(1,1:i);
       yNumerical=yNumerical(1,1:i); 
       break
    else 
       yNumerical(i+1)=yassume;
    end   
end
